#common functions
import ConfigParser
import logging
import sys
import socket
import commands
import subprocess, threading, datetime, os, time, signal
import re
from urlparse import *
import constant
import urllib2
import cutils

CONFIG_PATH = '../etc/config.ini'
is_metal_g = -1
is_monios_g = -1
is_metal_mutex = threading.Lock()
is_monios_mutex = threading.Lock()

def get_agent_name(config_path):
    config = ConfigParser.ConfigParser()
    config.read(config_path)
    return config.get('name','agent_name')
    
def get_agent_version(config_path):
    config = ConfigParser.ConfigParser()
    config.read(config_path)
    return config.get('name','agent_version')

def generate_config(config_path, section_name):
    config = ConfigParser.ConfigParser()
    config.read(config_path)
    config_dict = { }
    if config.has_section(section_name):
        for option_pair in config.items(section_name):
            config_dict[option_pair[0]] = option_pair[1]
    return config_dict

VM_METHON_DICT = {'/proc/cpuinfo':['hypervisor'],
    '/sys/devices/system/clocksource/clocksource0/available_clocksource':['kvm'],
    #in arm machine the value is cvm
    '/sys/class/dmi/id/product_name':['CVM','KVM']}
def is_metal():
    global is_metal_g
    global is_metal_mutex
    if is_metal_g != -1:
           return is_metal_g

    if is_metal_mutex.acquire():
        if is_metal_g != -1:
            is_metal_mutex.release()
            return is_metal_g
        for key in VM_METHON_DICT:
            if os.path.exists(key):
                with open(key) as fd:
                    temp =  fd.read()
                for key_word in VM_METHON_DICT[key]:
                    if(temp.find(key_word) != -1):
                        is_metal_g = 0
                        is_metal_mutex.release()
                        return is_metal_g
        is_metal_g = 1
        is_metal_mutex.release()
        return is_metal_g

def is_minios():
    global is_monios_g
    global is_monios_mutex
    if is_monios_g != -1:
        return is_monios_g

    if is_monios_mutex.acquire():
        if is_monios_g != -1:
            is_monios_mutex.release()
            return is_monios_g

        isminios = os.path.exists('/etc/minios_version')
        is_monios_g = is_metal() & isminios
	
    	is_monios_mutex.release()
    	return is_monios_g


def console_logger():
    logger = logging.getLogger()
    logger.addHandler(logging.StreamHandler(sys.stdout))
    logger.setLevel(logging.DEBUG)
    return logger


_local_ip =  ''
def local_ip():
    """Get local ip """
    global _local_ip
    if len(_local_ip)>0:
        return _local_ip
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect(('172.0.0.1', 80))
        _local_ip = sock.getsockname()[0]
    except Exception:
        getIpCmd = "ip route |grep -v eth0|awk '/src/{print $NF}'"
        _local_ip = commands.getoutput(getIpCmd).split("\n")[0]
    return _local_ip

class CommUtils :
    @staticmethod
    def ExecuteCmd(cmd):
        (status,output) = commands.getstatusoutput(cmd)
        if status == 0:
            return output
        return ""
    @staticmethod
    def ExecuteTimeoutCommand(command, timeout):
        """call shell-command and either return its output or kill it
        if it doesn't normally exit within timeout seconds and return None"""
        start = datetime.datetime.now()
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        while process.poll() is None:
            time.sleep(0.1)
            now = datetime.datetime.now()
            if (now - start).seconds> timeout:
                #os.kill(process.pid, signal.SIGKILL)
                #os.waitpid(-1, os.WNOHANG)
                '''modified by cloudyang 2014.07.15
                    fix for the fd-leak bug in subprocess module in python 2.6
                    hope it works'''
                process.kill()
                process.wait()
                return ""
        return process.stdout.read()

    @staticmethod
    def GetBlockDevices():
        try:
            f = open('/proc/partitions','r')
            l = f.readlines()
        except Exception,e:
            return
        n = 0
        res = []
        for x in l:
            n+=1
            if n <= 2:
               continue
            arr = x.split()
            chk_name = "/sys/class/block/" + arr[3]
            if os.path.exists(os.path.realpath(chk_name)):
                try_path = chk_name + "/dm/name"
                try :
                    f2 = open(try_path)
                    best_name = "/dev/mapper/" + f2.read()[0:-1]
                    f2.close()
                except :
                    best_name = "/dev/" + arr[3]
                res.append(best_name)
        f.close()
        return res

    @staticmethod
    def dns_parse(url):
        try:
            parse_url = urlparse(url)
            domain = parse_url.netloc
            ip = socket.gethostbyname(domain)
            pattern = "^(?:[0-9]{1,3}\.){3}[0-9]{1,3}$"
            if not re.match(pattern, ip):
                return url
            return url.replace(domain,ip)
        except Exception, e:
            return url

class incrementFileReadMng(object):
    def __init__(self, max_read_size=4096*1024, get_last_fn=None):
        self.max_read_size = max_read_size  # max size to read each time
        self.log_offset = -1
        self.log_size = 0
        self.fd = None
        self.data_arrived = True
        self.last = ''  # last file, may be name or inode
        self.current = ''
        self.current_size = 0
        self.get_last_fn = get_last_fn

    def open_last_file(self, path):
        log_size = os.path.getsize(path)
        # no need to read rotated data
        if log_size >= self.max_read_size:
            raise Exception('pass')
        # open path to avoid being rotated
        self.current = open(path, 'r')
        self.current_size = log_size
        left = self.max_read_size - log_size
        # get last file from self.last
        last_file = self.get_last_fn(self.last)
        last_size = os.path.getsize(last_file)
        offset = last_size - left
        if offset > self.log_offset:
            self.log_offset = offset
        self.fd = open(last_file, 'r')
        self.log_size = last_size

    # before open, path may be rotated, last was used to judge self.last for same file
    def inc_open_file(self, path, last=None):
        self.data_arrived = True
        # rotated
        if self.get_last_fn and self.last and last and last != self.last:
            try:
                self.open_last_file(path)
                self.last = last
                return True
            except Exception:
                self.inc_close_file()
        self.last = last
        try:
            self.log_size = os.path.getsize(path)

            # no data incremented
            if self.log_offset == self.log_size:
                self.data_arrived = False
                return False

            # first time run
            if self.log_offset == -1:
                self.log_offset = self.log_size
                return False

            # truncated or rotated
            if self.log_offset > self.log_size:
                self.log_offset = 0

            # only read max_read_size data
            if self.log_size - self.log_offset > self.max_read_size:
                self.log_offset = self.log_size - self.max_read_size
            self.fd = open(path, 'r')
        except Exception:
            self.inc_close_file()
            return False
        return True

    # must close file after each read
    def inc_close_file(self):
        if self.current:
            self.current.close()
            self.current = None
        if self.fd:
            self.fd.close()
            self.fd = None
        self.log_size = 0

    def open_current_file(self):
        if self.fd:
            self.fd.close()
        self.log_offset = 0
        self.fd = self.current
        self.current = None
        self.log_size = self.current_size

    def inc_seek_read_lines(self, size):
        if self.log_offset >= self.log_size:
            if not self.current:
                return ''
            self.open_current_file()
        self.fd.seek(self.log_offset, 0)
        buf = self.fd.read(size)
        if not buf and not self.current:
            return buf
        i = 1
        d_len = len(buf)
        if d_len < size and self.current:
            left = size - d_len
            self.open_current_file()
            data = self.fd.read(left)
            buf += data
            d_len = len(buf)
        # return data end with '\n' to avoid splitting data in lines
        while i <= d_len:
            if buf[-i] == '\n':
                break
            i += 1
        res = buf[0:d_len - i + 1]
        self.log_offset += d_len - i + 1
        return res

    # get last file if rotated
    def get_offload_last(self, last):
        return self.last

def check_net(url, time_out):
    try:
        urllib2.urlopen(url, timeout=time_out)
    except:
        return False
    return True

def get_url(server,new_url,old_url):
    get_config = generate_config(constant.PLUGIN_CONFIG_PATH,server)
    new_url = get_config[new_url]
    old_url = get_config[old_url]
    if check_net(new_url,3) is True:
        return new_url
    else:
        return old_url

'''
add com function get_vm_uuid
date:2021/09/18
author:kkingzhang
'''
vm_uuid = ''
metadata = ""
uuid_mutex = threading.Lock()
def get_vm_uuid():
    if is_minios():
        return ''
    global vm_uuid
    global uuid_mutex
    global metadata
    if uuid_mutex.acquire():
        if vm_uuid != '':
            uuid_mutex.release()
            return vm_uuid;
        if not is_metal():
            #get cvm uuid from dmi info
            uuid_serial = "/sys/class/dmi/id/product_serial"
            if os.path.exists(uuid_serial):
                with open(uuid_serial,'r') as fd:
                    first_char = fd.read(1)
                    fd.seek(0)
                    if not first_char.isalnum():
                        mount_file = "/etc/uuid"
                        if os.path.exists(mount_file):
                            with open('/etc/uuid', 'r') as f:
                                uuid = f.read().strip()
                            uuid_value = uuid.split('=')[-1].strip()
                            with open('/etc/uuid_to_serial', 'w') as f:
                                f.write(uuid_value)
                            mount_command = "mount --bind /etc/uuid_to_serial /sys/class/dmi/id/product_serial"
                            subprocess.Popen(mount_command, shell=True) 
                    vm_uuid = fd.readline().strip().lower()
                    uuid_mutex.release()
                    return vm_uuid
            uuid_path='/etc/uuid'
            if os.path.exists(uuid_path):
                with open(uuid_path,'r') as f:
                    vm_uuid = f.read().split('=')[1].replace('"','').replace('\'','').strip()
                    uuid_mutex.release()
                    return vm_uuid
        #we have to get baremetal uuid from metadata service
        if metadata == "":
            metadata = generate_config(constant.PLUGIN_CONFIG_PATH, "MetaDataServer")
        url = metadata['metadata_url'] + 'uuid'
        response = urllib2.urlopen(url, timeout=2)
        content = response.read()
        pattern = "^[0-9a-z]{8}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{4}-[0-9a-z]{12}$"
        if re.match(pattern, content):
            vm_uuid = content
            uuid_mutex.release()
            return vm_uuid
        vm_uuid = ""
        uuid_mutex.release()
        return vm_uuid

_client_key = ''
def get_client_key():
    CLIENT_KEY_PATH = '/usr/local/qcloud/monitor/config/clientKey.ini'
    global _client_key
    if len(_client_key) > 0:
        return _client_key
    try:
        if os.path.isfile(CLIENT_KEY_PATH):
            key_file = open(CLIENT_KEY_PATH, "r")
            lines = key_file.readlines()
            for line in lines:
                if line.startswith("clientKey") and line.find('=')>0 :
                    _client_key = line.split('=')[1].strip()
                    break
            key_file.close()
        else:
            import urllib2,json,time
            import constant
            url = get_url('UpdateServer','new_update_url','update_url')
            config = generate_config(constant.PLUGIN_CONFIG_PATH, 'UpdateServer')
            uuid = get_vm_uuid()
            req = {
                "version": "v1.0",
                "caller": "agent",
                "password": "",
                "callee": "autoupdate",
                "eventId": 101,
                "timestamp": time.time(),
                'interface': {
                    "interfaceName": "GetClientKey",
                    "para": {
                        "uuid": uuid
                    }
                }
            }
            http_ret = urllib2.urlopen(url, json.dumps(req), float(config['update_timeout']))
            response = http_ret.read()
            json_resp = json.loads(response)
            retcode = int(json_resp["returnCode"])
            if retcode != 0:
                print json_resp["returnMsg"]
            else: 
                _client_key = json_resp["returnData"]["clientKey"]
                dir, file = os.path.split(CLIENT_KEY_PATH)
                if not os.path.exists(dir):
                    os.makedirs(dir)
                with open(CLIENT_KEY_PATH, "w") as f:
                    str = 'clientKey=' + _client_key
                    f.writelines(str)
            
    except Exception,e:
        print e
    return _client_key

class MetadataSingleton(object):
    _instance = None
    _instance_lock = threading.Lock()

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with cls._instance_lock:
                if not cls._instance:
                    cls._instance = super(MetadataSingleton, cls).__new__(cls)
        return cls._instance

    def __init__(self, logger=None):
        self.instance_id = ''
        self.instance_id_mutex = threading.Lock()
        self.ins_id_fail_count = -1
        self.ins_id_last_fail_time = 0
        self.retry_intervals = [60, 600, 1800, 3600]
        self.app_id = ''
        self.app_id_mutex = threading.Lock()
        self.metadata = ''
        self.app_id_update_period = 3600
        self.app_id_last_update_time = 0
        self.app_id_fail_count = -1
        self.app_id_last_fail_time = 0
        self.logger = logger
        self.project_id = ""
        self.project_id_mutex = threading.Lock()
        self.project_id_update_period = 3600
        self.project_id_last_update_time = 0
        
    def get_vm_project_id(self):
        return 0

    def get_vm_instance_id(self):
        if self.instance_id != '':
            return self.instance_id
        if self.ins_id_last_fail_time != 0 and (int(time.time()) - self.ins_id_last_fail_time < self.retry_intervals[self.ins_id_fail_count]):
            return self.instance_id
        if self.metadata == '':
            self.metadata = generate_config(constant.PLUGIN_CONFIG_PATH, "MetaDataServer")

        if self.instance_id_mutex.acquire():
            if self.instance_id != '':
                self.instance_id_mutex.release()
                return self.instance_id
            try:
                url = self.metadata['metadata_url'] + 'instance-id'
                response = urllib2.urlopen(url, timeout=2)
                content = response.read()
                if len(content) > 0 and len(content.splitlines()) == 1:
                    self.instance_id = content
                    self.instance_id_mutex.release()
                    self.ins_id_last_fail_time = 0
                    return self.instance_id
            except Exception as e:
                self.instance_id = ''
                self.ins_id_last_fail_time = int(time.time())
                if self.ins_id_fail_count < len(self.retry_intervals) - 1:
                    self.ins_id_fail_count += 1
                self.logger.error("get instance_id from metadata fail,  %s ,after %s s request!" % (e, self.retry_intervals[self.ins_id_fail_count]))
            self.instance_id_mutex.release()
            return self.instance_id

    def get_vm_app_id(self):
        now = int(time.time())
        if self.app_id != '' and (now - self.app_id_last_update_time < self.app_id_update_period):
            return self.app_id
        if self.app_id_last_fail_time != 0 and (now - self.app_id_last_fail_time < self.retry_intervals[self.app_id_fail_count]):
            return self.app_id
        if self.metadata == '':
            self.metadata = generate_config(constant.PLUGIN_CONFIG_PATH, "MetaDataServer")
        if self.app_id_mutex.acquire():
            if self.app_id != '':
                self.app_id_mutex.release()
                return self.app_id
            try:
                url = self.metadata['metadata_url'] + 'app-id'
                response = urllib2.urlopen(url, timeout=2)
                content = response.read()
                if len(content) > 0 and len(content.splitlines()) == 1:
                    self.app_id = content
                    self.app_id_mutex.release()
                    self.app_id_last_update_time = int(time.time())
                    self.app_id_last_fail_time = 0
                    return self.app_id
            except Exception as e:
                self.app_id = ''
                self.app_id_last_fail_time = int(time.time())
                if self.app_id_fail_count < len(self.retry_intervals) - 1:
                    self.app_id_fail_count += 1
                self.logger.error("get app_id from metadata fail, %s ,after %s s request!" % (e, self.retry_intervals[self.ins_id_fail_count]))
            self.app_id_mutex.release()
            return self.app_id

if "__main__" == __name__:
    while True:
        print is_metal()
        print is_minios()
        print get_vm_uuid()
        metadata_instance = cutils.MetadataSingleton()
        instance_id = metadata_instance.get_vm_instance_id()
        app_id = metadata_instance.get_vm_app_id()
        project_id = metadata_instance.get_vm_project_id()
        time.sleep(5)
